// Copyright (c) Erik Stenborg, 2002. (E-mail: erik.stenborg@home.se)
// Licensed under LGPL: http://www.gnu.org/copyleft/lesser.html

unit OggTag;

interface

uses
  SysUtils, Classes;

type
  TVorbisInfo = packed record
    PacketType: Byte;
    Vorbis: array[0..5] of Char;
    VorbisVersion: Cardinal;
    Channels: Byte;
    SampleRate: Cardinal;
    MaxBitrate, NomBitrate, MinBitrate: Integer;
  end;

  TVorbisComment = record
    Comments: array of String;
    CommentLengths: array of Cardinal;
    CommentCount: Cardinal;
    Vendor: String;
    VendorLength: Cardinal;
  end;

  TOggTag = class
  private
    FLength: Int64;
    FVorbisSerialNr: Cardinal;
    FOldSize: Integer;
    function FSizeDifference: Integer;
    function GetCount: Integer;
    procedure SetCount(Value: Integer);
    function GetKey(Index: Integer): WideString;
    procedure SetKey(Index: Integer; Value: WideString);
    function GetValue(Index: WideString): WideString;
    procedure SetValue(Index: WideString; Value: WideString);
    function GetValueI(Index: Integer): WideString;
    procedure SetValueI(Index: Integer; Value: WideString);
    function LoadVI(S: TStream): Cardinal;
    function LoadVC(S: TStream): Cardinal;
    function GetLength(S: TStream): Int64;
    function GetDuration: Double;
  public
    VC: TVorbisComment;
    VI: TVorbisInfo;
    Padding: Integer;
    constructor Create;
    destructor Destroy; override;
    procedure LoadFromStream(S: TStream);
    procedure LoadFromFile(FN: String);
    procedure SaveToFile(FN: String);
    function GetIndexOf(Index: WideString): Integer;
    procedure DeleteI(Index: Integer);
    procedure Delete(Index: WideString);
    property Key[Index: Integer]: WideString read GetKey write SetKey;
    property Value[Index: WideString]: WideString read GetValue write SetValue;
    property ValueI[Index: Integer]: WideString read GetValueI write SetValueI;
    property Count: Integer read GetCount write SetCount;
    property LastGranule: Int64 read FLength;
    property SizeDifference: Integer read FSizeDifference;
    property Duration: Double read GetDuration;
  end;

  EOggStream = class(Exception)
  end;

  EOggChecksum = class(Exception)
  end;

implementation

type
  PCardinal = ^Cardinal;
  PByte = ^Byte;
  TByteArray = packed array of Byte;
  TCharArray = packed array of Char;
  TSegmentTable = array[0..254] of Byte;

  TOggPageHeader = packed record
    CapturePattern: packed array [1..4] of Char;
    StreamStructureVersion,
    HeaderTypeFlag: Byte;
    AbsolutePosition: Int64;
    StreamSerialNumber,
    PageSequenceNumber,
    Checksum: Cardinal;
    PageSegments: Byte;
    SegmentTable: TSegmentTable;
    CurrentSegment: Byte;
    PageBodyLength: Integer;
  end;

  TOggPageWriter = class
  public
    constructor Create(S: TStream);
    destructor Destroy; override;
    procedure Clear;
    procedure Add(Packet: TMemoryStream; Skip: Integer);
    procedure Flush;
  protected
    Header: TOggPageHeader;
  private
    Data: TMemoryStream;
    OutStream: TStream;
  end;

//----------------------------------------------------------------------------------------------------------------------
// UTF8-functions from JCL Unicode library

  // Unicode transformation formats (UTF) data types
  UTF7 = Char;
  UTF8 = Char;
  UTF16 = WideChar;
  UTF32 = Cardinal;

  // UTF conversion schemes (UCS) data types
  PUCS4 = ^UCS4;
  UCS4 = Cardinal;
  PUCS2 = PWideChar;
  UCS2 = WideChar;

const
  ReplacementCharacter: UCS4 = $0000FFFD;
  MaximumUCS2: UCS4 = $0000FFFF;
  MaximumUTF16: UCS4 = $0010FFFF;
  MaximumUCS4: UCS4 = $7FFFFFFF;

  SurrogateHighStart: UCS4 = $D800;
  SurrogateHighEnd: UCS4 = $DBFF;
  SurrogateLowStart: UCS4 = $DC00;
  SurrogateLowEnd: UCS4 = $DFFF;

  halfShift: Integer = 10;

  halfBase: UCS4 = $0010000;
  halfMask: UCS4 = $3FF;

  offsetsFromUTF8: array[0..5] of UCS4 = ($00000000, $00003080, $000E2080, $03C82080, $FA082080, $82082080);

  bytesFromUTF8: array[0..255] of Byte = (
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
	2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2, 3,3,3,3,3,3,3,3,4,4,4,4,5,5,5,5);

  firstByteMark: array[0..6] of Byte = ($00, $00, $C0, $E0, $F0, $F8, $FC);

function WideStringToUTF8(S: WideString): AnsiString;
var
  ch: UCS4;
  L, J, T,
  bytesToWrite: Word;
  byteMask: UCS4;
  byteMark: UCS4;
begin
  if Length(S) = 0 then
  begin
    Result := '';
    Exit;
  end;
  SetLength(Result, Length(S) * 6); // assume worst case
  T := 1;
  for J := 1 to Length(S) do
  begin
    byteMask := $BF;
    byteMark := $80;
    ch := UCS4(S[J]);
    if ch < $80 then
      bytesToWrite := 1
    else
    if ch < $800 then
      bytesToWrite := 2
    else
    if ch < $10000 then
      bytesToWrite := 3
    else
    if ch < $200000 then
      bytesToWrite := 4
    else
    if ch < $4000000 then
      bytesToWrite := 5
    else
    if ch <= MaximumUCS4 then
      bytesToWrite := 6
    else
    begin
      bytesToWrite := 2;
      ch := ReplacementCharacter;
    end;
    for L := bytesToWrite downto 2 do
    begin
      Result[T + L - 1] := Char((ch or byteMark) and byteMask);
      ch := ch shr 6;
    end;
    Result[T] := Char(ch or firstByteMark[bytesToWrite]);
    Inc(T, bytesToWrite);
  end;
  SetLength(Result, T - 1); // assume worst case
end;

//----------------------------------------------------------------------------------------------------------------------

function UTF8ToWideString(S: AnsiString): WideString;
var
  L, J, T: Cardinal;
  ch: UCS4;
  extraBytesToWrite: Word;
begin
  if Length(S) = 0 then
  begin
    Result := '';
    Exit;
  end;
  SetLength(Result, Length(S)); // create enough room
  L := 1;
  T := 1;
  while L <= Cardinal(Length(S)) do
  begin
    ch := 0;
    extraBytesToWrite := bytesFromUTF8[Ord(S[L])];
    for J := extraBytesToWrite downto 1 do
    begin
      ch := ch + Ord(S[L]);
      Inc(L);
      ch := ch shl 6;
    end;
    ch := ch + Ord(S[L]);
    Inc(L);
    ch := ch - offsetsFromUTF8[extraBytesToWrite];
    if ch <= MaximumUCS2 then
    begin
      Result[T] := WideChar(ch);
      Inc(T);
    end
    else
    if ch > MaximumUCS4 then
    begin
      Result[T] := WideChar(ReplacementCharacter);
      Inc(T);
    end
    else
    begin
      ch := ch - halfBase;
      Result[T] := WideChar((ch shr halfShift) + SurrogateHighStart);
      Inc(T);
      Result[T] := WideChar((ch and halfMask) + SurrogateLowStart);
      Inc(T);
    end;
  end;
  SetLength(Result, T - 1); // now fix up length
end;

//----------------------------------------------------------------------------------------------------------------------

const
  CRCTab: array[0..255] of Cardinal = (
  $00000000,$04c11db7,$09823b6e,$0d4326d9,
  $130476dc,$17c56b6b,$1a864db2,$1e475005,
  $2608edb8,$22c9f00f,$2f8ad6d6,$2b4bcb61,
  $350c9b64,$31cd86d3,$3c8ea00a,$384fbdbd,
  $4c11db70,$48d0c6c7,$4593e01e,$4152fda9,
  $5f15adac,$5bd4b01b,$569796c2,$52568b75,
  $6a1936c8,$6ed82b7f,$639b0da6,$675a1011,
  $791d4014,$7ddc5da3,$709f7b7a,$745e66cd,
  $9823b6e0,$9ce2ab57,$91a18d8e,$95609039,
  $8b27c03c,$8fe6dd8b,$82a5fb52,$8664e6e5,
  $be2b5b58,$baea46ef,$b7a96036,$b3687d81,
  $ad2f2d84,$a9ee3033,$a4ad16ea,$a06c0b5d,
  $d4326d90,$d0f37027,$ddb056fe,$d9714b49,
  $c7361b4c,$c3f706fb,$ceb42022,$ca753d95,
  $f23a8028,$f6fb9d9f,$fbb8bb46,$ff79a6f1,
  $e13ef6f4,$e5ffeb43,$e8bccd9a,$ec7dd02d,
  $34867077,$30476dc0,$3d044b19,$39c556ae,
  $278206ab,$23431b1c,$2e003dc5,$2ac12072,
  $128e9dcf,$164f8078,$1b0ca6a1,$1fcdbb16,
  $018aeb13,$054bf6a4,$0808d07d,$0cc9cdca,
  $7897ab07,$7c56b6b0,$71159069,$75d48dde,
  $6b93dddb,$6f52c06c,$6211e6b5,$66d0fb02,
  $5e9f46bf,$5a5e5b08,$571d7dd1,$53dc6066,
  $4d9b3063,$495a2dd4,$44190b0d,$40d816ba,
  $aca5c697,$a864db20,$a527fdf9,$a1e6e04e,
  $bfa1b04b,$bb60adfc,$b6238b25,$b2e29692,
  $8aad2b2f,$8e6c3698,$832f1041,$87ee0df6,
  $99a95df3,$9d684044,$902b669d,$94ea7b2a,
  $e0b41de7,$e4750050,$e9362689,$edf73b3e,
  $f3b06b3b,$f771768c,$fa325055,$fef34de2,
  $c6bcf05f,$c27dede8,$cf3ecb31,$cbffd686,
  $d5b88683,$d1799b34,$dc3abded,$d8fba05a,
  $690ce0ee,$6dcdfd59,$608edb80,$644fc637,
  $7a089632,$7ec98b85,$738aad5c,$774bb0eb,
  $4f040d56,$4bc510e1,$46863638,$42472b8f,
  $5c007b8a,$58c1663d,$558240e4,$51435d53,
  $251d3b9e,$21dc2629,$2c9f00f0,$285e1d47,
  $36194d42,$32d850f5,$3f9b762c,$3b5a6b9b,
  $0315d626,$07d4cb91,$0a97ed48,$0e56f0ff,
  $1011a0fa,$14d0bd4d,$19939b94,$1d528623,
  $f12f560e,$f5ee4bb9,$f8ad6d60,$fc6c70d7,
  $e22b20d2,$e6ea3d65,$eba91bbc,$ef68060b,
  $d727bbb6,$d3e6a601,$dea580d8,$da649d6f,
  $c423cd6a,$c0e2d0dd,$cda1f604,$c960ebb3,
  $bd3e8d7e,$b9ff90c9,$b4bcb610,$b07daba7,
  $ae3afba2,$aafbe615,$a7b8c0cc,$a379dd7b,
  $9b3660c6,$9ff77d71,$92b45ba8,$9675461f,
  $8832161a,$8cf30bad,$81b02d74,$857130c3,
  $5d8a9099,$594b8d2e,$5408abf7,$50c9b640,
  $4e8ee645,$4a4ffbf2,$470cdd2b,$43cdc09c,
  $7b827d21,$7f436096,$7200464f,$76c15bf8,
  $68860bfd,$6c47164a,$61043093,$65c52d24,
  $119b4be9,$155a565e,$18197087,$1cd86d30,
  $029f3d35,$065e2082,$0b1d065b,$0fdc1bec,
  $3793a651,$3352bbe6,$3e119d3f,$3ad08088,
  $2497d08d,$2056cd3a,$2d15ebe3,$29d4f654,
  $c5a92679,$c1683bce,$cc2b1d17,$c8ea00a0,
  $d6ad50a5,$d26c4d12,$df2f6bcb,$dbee767c,
  $e3a1cbc1,$e760d676,$ea23f0af,$eee2ed18,
  $f0a5bd1d,$f464a0aa,$f9278673,$fde69bc4,
  $89b8fd09,$8d79e0be,$803ac667,$84fbdbd0,
  $9abc8bd5,$9e7d9662,$933eb0bb,$97ffad0c,
  $afb010b1,$ab710d06,$a6322bdf,$a2f33668,
  $bcb4666d,$b8757bda,$b5365d03,$b1f740b4);

procedure ChecksumCalc(var crc: Cardinal; const buffer; size: Cardinal);
var
  i: cardinal;
  b: array[0..MaxInt - 1] of Byte absolute buffer;
begin
  for i := 0 to size-1 do
    crc := (crc shl 8) xor CRCTab[((crc shr 24) and $ff) xor b[i]];
end;

function SumByte(Data: TSegmentTable; Len: Byte): Integer;
var
  I: Integer;
begin
  Result := 0;
  for I := 0 to Len-1 do Inc(Result, Data[I]);
end;

procedure ReadPageHeader(var PageHeader: TOggPageHeader; Stream: TStream);
begin
  FillChar(PageHeader, SizeOf(PageHeader), 0);
  Stream.ReadBuffer(PageHeader, 27);
  if PageHeader.CapturePattern <> 'OggS' then
    raise EOggStream.Create('Not a valid Ogg stream');
  if PageHeader.StreamStructureVersion <> 0 then
    raise EOggStream.CreateFmt('Not a valid OGG stream structure version: %d', [PageHeader.StreamStructureVersion]);
  Stream.ReadBuffer(PageHeader.SegmentTable[0], PageHeader.PageSegments);
  PageHeader.PageBodyLength := SumByte(PageHeader.SegmentTable, PageHeader.PageSegments);
end;

procedure WriteNewCRC(S: TStream);
var
  Header: TOggPageHeader;
  StartPos, CRC: Cardinal;
  B: array[0..255] of Byte;
  I: Integer;
begin
  StartPos := S.Position;
  ReadPageHeader(Header, S);
  Header.Checksum := 0;
  CRC := 0;
  ChecksumCalc(CRC, Header, 27);
  ChecksumCalc(CRC, Header.SegmentTable[0], Header.PageSegments);
  for I := 1 to Header.PageBodyLength div SizeOf(B) do begin
    S.ReadBuffer(B[0], SizeOf(B));
    ChecksumCalc(CRC, B[0], SizeOf(B));
  end;
  S.ReadBuffer(B[0], Header.PageBodyLength mod SizeOf(B));
  ChecksumCalc(CRC, B[0], Header.PageBodyLength mod SizeOf(B));
  S.Seek(StartPos + 22, soFromBeginning);
  S.WriteBuffer(CRC, 4);
  S.Seek(StartPos, soFromBeginning);
end;

function CommentSize(VC: TVorbisComment): Integer;
var
  I: Integer;
begin
  Result := 7 + 4 + Cardinal(Length(VC.Vendor)) + 4 + VC.CommentCount * 4 + 1;
  for I := 0 to VC.CommentCount-1 do
    Inc(Result, Length(VC.Comments[I]));
end;

function PackComment(const VC: TVorbisComment; var S: TMemoryStream): Integer;
const
  VorbisString: array[0..6] of Char = #3'vorbis';
var
  N, I: Integer;
begin
  S.Clear;
  S.WriteBuffer(VorbisString[0], 7);
  N := Length(VC.Vendor);
  S.WriteBuffer(N, 4);
  S.WriteBuffer(VC.Vendor[1], N);
  S.WriteBuffer(VC.CommentCount, 4);
  for I := 0 to VC.CommentCount-1 do begin
    N := Length(VC.Comments[I]);
    S.WriteBuffer(N, 4);
    S.WriteBuffer(VC.Comments[I,1], N);
  end;
  N := 1;
  S.WriteBuffer(N, 1); //framing bit
  Result := S.Position;
  S.Seek(0, soFromBeginning);
end;

function Sync(S: TStream): Boolean;
var
  C: array[1..5] of Char;
begin
  Result := True;
  try
    repeat
      repeat S.ReadBuffer(C[1], 1) until C[1] = 'O';
      S.ReadBuffer(C[2], 4);  //Just peek at the next 4 bytes
      S.Seek(-4, soFromCurrent);
    until C = 'OggS'#0;
    S.Seek(-1, soFromCurrent);
  except
    //Suddenly went past the end (truncated stream)
    on EReadError do Result := False;
  end;
end;

function ReadPacket(var Packet: TMemoryStream; var Head: TOggPageHeader; var CRC: Cardinal; const S: TStream): Cardinal;
var
  StreamNr: Cardinal;
  I: Integer;
begin
  Packet.Clear;
  StreamNr := Head.StreamSerialNumber;
  while True do begin
    for I := Head.CurrentSegment to Head.PageSegments-1 do begin
      Packet.CopyFrom(S, Head.SegmentTable[I]);
      if Head.SegmentTable[I] < 255 then begin
        Result := Packet.Position;
        Packet.Seek(0, soFromBeginning);
        ChecksumCalc(CRC, Packet.Memory^, Packet.Size);
        Head.CurrentSegment := I + 1;
        Exit;
      end;
    end;
    if CRC <> Head.Checksum then
      raise EOggChecksum.CreateFmt('Checksum mismatch. Should be: %d. Actual: %d.', [Head.Checksum, CRC]);
    ReadPageHeader(Head, S);
    while Head.StreamSerialNumber <> StreamNr do begin
      S.Seek(Head.PageBodyLength, soFromCurrent);
      ReadPageHeader(Head, S);
    end;
    CRC := 0;
    ChecksumCalc(CRC, Head, 27);
    ChecksumCalc(CRC, Head.SegmentTable[0], Head.PageSegments);
  end;
end;

function TOggTag.LoadVI(S: TStream): Cardinal;
var
  PageHeader: TOggPageHeader;
  Packet: TMemoryStream;
  VorbisInfo: TVorbisInfo;
  HasVorbisStream: Boolean;
  CRC: Cardinal;
begin
  HasVorbisStream := False;
  FVorbisSerialNr := Cardinal(-1);
  Packet := TMemoryStream.Create;
  CRC := 0;
  if Sync(s) then repeat
    ReadPageHeader(PageHeader, S);
    ReadPacket(Packet, PageHeader, CRC, S);
    Packet.Read(VorbisInfo, SizeOf(VorbisInfo));
    if (VorbisInfo.PacketType = 1) and
       (VorbisInfo.Vorbis = 'vorbis') and
       (VorbisInfo.VorbisVersion = 0) then begin
      HasVorbisStream := True;
      FVorbisSerialNr := PageHeader.StreamSerialNumber;
      VI := VorbisInfo;
    end;
  until HasVorbisStream or (PageHeader.HeaderTypeFlag <> 2);
  if not HasVorbisStream then
    raise EOggStream.Create('The stream does not contain Vorbis data.');
  Packet.Free;
  Result := S.Position;
end;

function TOggTag.LoadVC(S: TStream): Cardinal;
var
  PageHeader: TOggPageHeader;
  Packet: TMemoryStream;
  VorbisInfo: TVorbisInfo;
  J: Integer;
  CRC: Cardinal;
begin
  ReadPageHeader(PageHeader, S);
  while PageHeader.StreamSerialNumber <> FVorbisSerialNr do begin
    S.Seek(PageHeader.PageBodyLength, soFromCurrent);
    ReadPageHeader(PageHeader, S);
  end;
  J := S.Position;
  CRC := 0;
  Packet := TMemoryStream.Create;
  Result := ReadPacket(Packet, PageHeader, CRC, S);
  FOldSize := S.Position - J;
  Packet.Read(VorbisInfo, 7);
  if (VorbisInfo.PacketType <> 3) or
     (VorbisInfo.Vorbis <> 'vorbis') then
    raise EOggStream.Create('The stream does not contain a Vorbis comment block.');
  Packet.ReadBuffer(VC.VendorLength, SizeOf(VC.VendorLength));
  SetLength(VC.Vendor, VC.VendorLength);
  Packet.ReadBuffer(VC.Vendor[1], VC.VendorLength);
  Packet.ReadBuffer(VC.CommentCount, SizeOf(VC.CommentCount));
  SetLength(VC.Comments, VC.CommentCount);
  SetLength(VC.CommentLengths, VC.CommentCount);
  for J := 0 to VC.CommentCount-1 do begin
    Packet.ReadBuffer(VC.CommentLengths[J], SizeOf(VC.CommentLengths[J]));
    SetLength(VC.Comments[J], VC.CommentLengths[J]);
    Packet.ReadBuffer(VC.Comments[J,1], VC.CommentLengths[J]);
  end;
  Packet.Free;
end;

function TOggTag.GetLength(S: TStream): Int64;
var
  Header: TOggPageHeader;
begin
  S.Seek(-60000, soFromEnd);
  Result := 0;
  if Sync(S) then repeat
    ReadPageHeader(Header, S);
    if Header.StreamSerialNumber = FVorbisSerialNr then Result := Header.AbsolutePosition;
    S.Seek(Header.PageBodyLength, soFromCurrent);
  until (Header.StreamSerialNumber = FVorbisSerialNr) and (Header.HeaderTypeFlag and 4 <> 0);
end;

procedure SegmentPacket(var Segments: Byte; var SegmentTable: TSegmentTable; DataSize: Cardinal; Finished: Boolean);
var
  NewSegments: Cardinal;
begin
  NewSegments := DataSize div 255;
  FillChar(SegmentTable[Segments], NewSegments, 255);
  Inc(Segments, NewSegments);
  if Finished then begin
    SegmentTable[Segments] := DataSize mod 255;
    Inc(Segments, 1);
  end;
end;

constructor TOggPageWriter.Create(S: TStream);
begin
  inherited Create;
  Data := TMemoryStream.Create;
  OutStream := S;
  Clear;
end;

destructor TOggPageWriter.Destroy;
begin
  Flush;
  Data.Free;
  inherited Destroy;
end;

procedure TOggPageWriter.Clear;
begin
  Data.Clear;
  FillChar(Header, SizeOf(Header), 0);
  Header.CapturePattern := 'OggS';
end;

procedure TOggPageWriter.Add(Packet: TMemoryStream; Skip: Integer);
begin
  SegmentPacket(Header.PageSegments, Header.SegmentTable, Packet.Size, true);
  Inc(Header.PageBodyLength, Packet.Size);
  Packet.Seek(Skip, soFromBeginning);
  Data.CopyFrom(Packet, Packet.Size - Skip);
end;

procedure TOggPageWriter.Flush;
var
  CRC: Cardinal;
begin
  if Header.PageBodyLength > 0 then begin
    Header.Checksum := 0;
    CRC := 0;
    ChecksumCalc(CRC, Header, 27);
    ChecksumCalc(CRC, Header.SegmentTable[0], Header.PageSegments);
    ChecksumCalc(CRC, Data.Memory^, Data.Size);
    Header.Checksum := CRC;
    OutStream.WriteBuffer(Header, 27);
    OutStream.WriteBuffer(Header.SegmentTable[0], Header.PageSegments);
    Data.Seek(0, soFromBeginning);
    OutStream.CopyFrom(Data, Data.Size);
    //prepare next page header
    Data.Clear;
    FillChar(Header.SegmentTable[0], 255, 0);
    Header.PageSegments := 0;
    Header.PageBodyLength := 0;
    Header.Checksum := 0;
    Inc(Header.PageSequenceNumber);
  end;
end;

procedure TOggTag.LoadFromStream(S: TStream);
begin
  LoadVI(S);
  LoadVC(S);
  FLength := -1;
end;

procedure TOggTag.LoadFromFile(FN: String);
var
  S: TFileStream;
begin
  S := TFileStream.Create(FN, fmOpenRead);
  S.Seek(0, soFromBeginning);
  LoadFromStream(S);
  FLength := GetLength(S);
  S.Free;
end;

procedure TOggTag.SaveToFile(FN: String);
var
  Src, Dest: TFileStream;
  CommentPageStart, AudioPageStart, NewSize, Room, Temp, PageNr: Integer;
  Packet2, Packet3: TMemoryStream;
  Header: TOggPageHeader;
  PageWriter: TOggPageWriter;
  CRC: Cardinal;
begin
  Src := TFileStream.Create(FN, fmOpenReadWrite);
  Src.Seek(0, soFromBeginning);
  LoadVI(Src);
  CommentPageStart := Src.Position;
  ReadPageHeader(Header, Src);
  PageNr := Header.PageSequenceNumber;
  Packet2 := TMemoryStream.Create;
  CRC := 0;
  Room := ReadPacket(Packet2, Header, CRC, Src);
  NewSize := PackComment(VC, Packet2);
  if NewSize <= Room then begin
    //pad with zeroes to old size
    Packet2.SetSize(Room);
    FillChar((PChar(Packet2.Memory)+NewSize)^, Room - NewSize, 0);
    Packet2.Seek(0, soFromBeginning);
    Src.Seek(CommentPageStart, soFromBeginning);
    ReadPageHeader(Header, Src);
    while Header.PageBodyLength < Room do begin
      Src.CopyFrom(Packet2, Header.PageBodyLength);
      Temp := Src.Position;
      Src.Seek(CommentPageStart, soFromBeginning);
      WriteNewCRC(Src);
      Dec(Room, Header.PageBodyLength);
      Src.Seek(Temp, soFromBeginning);
      CommentPageStart := Temp;
      ReadPageHeader(Header, Src);
    end;
    Src.CopyFrom(Packet2, Room);
    Src.Seek(CommentPageStart, soFromBeginning);
    WriteNewCRC(Src);
    Src.Free;
  end else begin
    Dest := TFileStream.Create(FN + '.temp', fmCreate);
    Packet2.SetSize(NewSize + Padding);
    FillChar((PChar(Packet2.Memory)+NewSize)^, Padding, 0);
    Packet3 := TMemoryStream.Create;
    ReadPacket(Packet3, Header, CRC, Src);
    AudioPageStart := Src.Position;
    Src.Seek(0, soFromBeginning);
    Dest.CopyFrom(Src, CommentPageStart);
    PageWriter := TOggPageWriter.Create(Dest);
    PageWriter.Header.StreamSerialNumber := Header.StreamSerialNumber;
    PageWriter.Header.PageSequenceNumber := PageNr;
    PageWriter.Add(Packet2, 0);
    PageWriter.Add(Packet3, 0);
    PageWriter.Flush;
    PageWriter.Free;
    Src.Seek(AudioPageStart, soFromBeginning);
    Dest.CopyFrom(Src, Src.Size - Src.Position);
    Dest.Free;
    Src.Free;
    DeleteFile(FN);
    RenameFile(FN + '.temp', FN);
    Packet3.Free;
  end;
  Packet2.Free;
end;


function TOggTag.GetKey(Index: Integer): WideString;
begin
  Result := WideString(Copy(VC.Comments[Index], 1, Pos('=', VC.Comments[Index]) - 1));
end;

procedure TOggTag.SetKey(Index: Integer; Value: WideString);
begin
  if Value <> '' then
    VC.Comments[Index] := UpperCase(Value) +
                          Copy(VC.Comments[Index], Pos('=', VC.Comments[Index]), MaxInt)
  else
    DeleteI(Index);
end;

function TOggTag.GetValue(Index: WideString): WideString;
var
  n: Integer;
begin
  n := GetIndexOf(Index);
  if n <> -1 then Result := GetValueI(n)
  else Result := '';
end;

procedure TOggTag.SetValue(Index: WideString; Value: WideString);
var
  n: Integer;
begin
  if Index <> '' then begin
    n := GetIndexOf(Index);
    if n <> -1 then SetValueI(n, Value)
    else if Value <> '' then begin
      Count := Count + 1;
      VC.Comments[Count-1] := UpperCase(Index) + '=' + WideStringToUTF8(Value);
    end;
  end;
end;

function TOggTag.GetValueI(Index: Integer): WideString;
begin
  Result := UTF8ToWideString(Copy(VC.Comments[Index], 1 + Pos('=', VC.Comments[Index]), MaxInt));
end;

procedure TOggTag.SetValueI(Index: Integer; Value: WideString);
begin
  if Value <> '' then
    VC.Comments[Index] := Copy(VC.Comments[Index], 1, Pos('=', VC.Comments[Index])) +
                          WideStringToUTF8(Value)
  else
    DeleteI(Index);
end;

function TOggTag.GetCount: Integer;
begin
  Result := VC.CommentCount;
end;

procedure TOggTag.SetCount(Value: Integer);
var
  I: Integer;
begin
  if Length(VC.Comments) < Value then SetLength(VC.Comments, 2*Value);
  for I := VC.CommentCount to Value-1 do VC.Comments[I] := '';
  VC.CommentCount := Value;
end;

function TOggTag.GetIndexOf(Index: WideString): Integer;
var
  S: String;
begin
  S := UpperCase(Index);
  for Result := 0 to Count-1 do
    if UpperCase(Key[Result]) = S then Exit;
  Result := -1;
end;

procedure TOggTag.DeleteI(Index: Integer);
begin
  if (Index >= 0) and (Index < Count) then begin
    VC.Comments[Index] := VC.Comments[Count-1];
    Count := Count - 1;
  end;
end;

procedure TOggTag.Delete(Index: WideString);
begin
  DeleteI(GetIndexOf(Index));
end;

constructor TOggTag.Create;
begin
  inherited Create;
  Padding := 500;
end;

function TOggTag.FSizeDifference: Integer;
begin
  Result := CommentSize(VC) - FOldSize;
end;

destructor TOggTag.Destroy;
begin
  SetLength(VC.Comments, 0);
  SetLength(VC.CommentLengths, 0);
  inherited Destroy;
end;

function TOggTag.GetDuration: Double;
begin
  Result := LastGranule / VI.SampleRate;
end;

end.
